/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>
#include <vconf.h>
#include <vconf-keys.h>
#include <limits.h>
#include <libgdbus/dbus-system.h>

#include "dd-deviced.h"
#include "deviced-priv.h"
#include "log.h"

#define PREDEF_PWROFF_POPUP			"pwroff-popup"
#define PREDEF_ENTERSLEEP			"entersleep"
#define PREDEF_LEAVESLEEP			"leavesleep"
#define PREDEF_REBOOT				"reboot"
#define PREDEF_BACKGRD				"backgrd"
#define PREDEF_FOREGRD				"foregrd"
#define PREDEF_ACTIVE				"active"
#define PREDEF_INACTIVE				"inactive"
#define PREDEF_SET_DATETIME			"set_datetime"
#define PREDEF_SET_TIMEZONE			"set_timezone"

#define PREDEF_SET_MAX_FREQUENCY		"set_max_frequency"
#define PREDEF_SET_MIN_FREQUENCY		"set_min_frequency"
#define PREDEF_RELEASE_MAX_FREQUENCY		"release_max_frequency"
#define PREDEF_RELEASE_MIN_FREQUENCY		"release_min_frequency"

#define ALARM_BUS_NAME		"com.samsung.alarm.manager"
#define ALARM_PATH_NAME		"/com/samsung/alarm/manager"
#define ALARM_INTERFACE_NAME	ALARM_BUS_NAME
#define ALARM_SET_TIME_METHOD	"alarm_set_time"

static int dbus_proc_handler(char* type, char *buf)
{
	GVariant *msg;
	int val;

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
						DEVICED_PATH_PROCESS,
						DEVICED_INTERFACE_PROCESS,
						type,
						g_variant_new("(sis)", type, 1, buf));
	if (!msg)
		return -EBADMSG;

	if (!dh_get_param_from_var(msg, "(i)", &val)) {
		_E("Failed to get signature(%s): no message", g_variant_get_type_string(msg));
		val = -EBADMSG;
		goto out;
	}

	_D("%s-%s: %d", DEVICED_INTERFACE_PROCESS, type, val);

out:
	g_variant_unref(msg);
	return val;
}

API int deviced_inform_foregrd(void)
{
	char buf[255];
	snprintf(buf, sizeof(buf), "%d", getpid());
	return dbus_proc_handler(PREDEF_FOREGRD, buf);
}

API int deviced_inform_backgrd(void)
{
	char buf[255];
	snprintf(buf, sizeof(buf), "%d", getpid());
	return dbus_proc_handler(PREDEF_BACKGRD, buf);
}

API int deviced_inform_active(pid_t pid)
{
	char buf[255];
	snprintf(buf, sizeof(buf), "%d", pid);
	return dbus_proc_handler(PREDEF_ACTIVE, buf);
}

API int deviced_inform_inactive(pid_t pid)
{
	char buf[255];
	snprintf(buf, sizeof(buf), "%d", pid);
	return dbus_proc_handler(PREDEF_INACTIVE, buf);
}

static int dbus_power_handler(char* type)
{
	GVariant *msg;
	int val;

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
						DEVICED_PATH_POWEROFF,
						DEVICED_INTERFACE_POWEROFF,
						"PowerOff",
						g_variant_new("(s)", type));
	if (!msg)
		return -EBADMSG;

	if (!dh_get_param_from_var(msg, "(i)", &val)) {
		_E("Failed to get signature(%s): no message", g_variant_get_type_string(msg));
		val = -EBADMSG;
		goto out;
	}

	_D("%s-%s: %d", DEVICED_INTERFACE_POWER, type, val);

out:
	g_variant_unref(msg);

	return val;
}

API int deviced_request_poweroff(void)
{
	return dbus_power_handler(PREDEF_PWROFF_POPUP);
}

API int deviced_request_reboot(void)
{
	return dbus_power_handler(PREDEF_REBOOT);
}

static int dbus_time_handler(char* type, char* buf)
{
	GVariant *msg;
	pid_t pid;
	char name[PATH_MAX];
	int ret, val;

	pid = getpid();
	ret = deviced_get_cmdline_name(pid, name, sizeof(name));
	if (ret != 0)
		snprintf(name, sizeof(name), "%d", pid);

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
						DEVICED_PATH_SYSNOTI,
						DEVICED_INTERFACE_SYSNOTI,
						type,
						g_variant_new("(sis)", type, 1, buf));
	if (!msg)
		return -EBADMSG;

	if (!dh_get_param_from_var(msg, "(i)", &val)) {
		_E("Failed to get signature(%s): no message", g_variant_get_type_string(msg));
		val = -EBADMSG;
		goto out;
	}

	_SI("[%s] %s-%s(%s): %d", name, DEVICED_INTERFACE_SYSNOTI, type, buf, val);
out:
	g_variant_unref(msg);

	return val;
}

static int alarm_set_time(time_t timet)
{
	GVariant *msg;
	pid_t pid;
	char name[PATH_MAX];
	int ret, val;

	pid = getpid();
	ret = deviced_get_cmdline_name(pid, name, sizeof(name));
	if (ret != 0)
		snprintf(name, sizeof(name), "%d", pid);
	_SI("[%s]start %s %ld.", name, ALARM_INTERFACE_NAME, timet);

	msg = dbus_handle_method_sync_with_reply_var(ALARM_BUS_NAME,
						ALARM_PATH_NAME,
						ALARM_INTERFACE_NAME,
						ALARM_SET_TIME_METHOD,
						g_variant_new("(i)", timet));
	if (!msg) {
		_E("Failed to call dbus_message_new_method_call(%s:%s-%s).",
			ALARM_PATH_NAME, ALARM_INTERFACE_NAME, ALARM_SET_TIME_METHOD);
		return -EBADMSG;
	}

	if (!dh_get_param_from_var(msg, "(i)", &val)) {
		_E("Failed to get type(%s): no message", g_variant_get_type_string(msg));
		val = -EBADMSG;
	}

	_SI("[%s]End %s %ld: %d", name, ALARM_INTERFACE_NAME, timet, val);

	g_variant_unref(msg);

	return val;
}

API int deviced_set_datetime(time_t timet)
{
	if (timet < 0L)
		return -1;
	return alarm_set_time(timet);
}

API int deviced_set_timezone(char *tzpath_str)
{
	if (tzpath_str == NULL)
		return -1;
	char buf[255];
	snprintf(buf, sizeof(buf), "%s", tzpath_str);
	return dbus_time_handler(PREDEF_SET_TIMEZONE, buf);
}

static int dbus_cpu_handler(char* type, char* buf_pid, char* buf_freq)
{
	GVariant *msg;
	int val;

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
						DEVICED_PATH_SYSNOTI,
						DEVICED_INTERFACE_SYSNOTI,
						type,
						g_variant_new("(siss)", type, 2, buf_pid, buf_freq));
	if (!msg)
		return -EBADMSG;

	if (!dh_get_param_from_var(msg, "(i)", &val)) {
		_E("Failed to get signature(%s): no message", g_variant_get_type_string(msg));
		val = -EBADMSG;
		goto out;
	}

	_D("%s-%s: %d", DEVICED_INTERFACE_SYSNOTI, type, val);
out:
	g_variant_unref(msg);
	return val;
}

API int deviced_request_set_cpu_max_frequency(int val)
{
	char buf_pid[8];
	char buf_freq[256];

	// to do - need to check new frequncy is valid
	snprintf(buf_pid, sizeof(buf_pid), "%d", getpid());
	snprintf(buf_freq, sizeof(buf_freq), "%d", val * 1000);

	return dbus_cpu_handler(PREDEF_SET_MAX_FREQUENCY, buf_pid, buf_freq);
}

API int deviced_request_set_cpu_min_frequency(int val)
{
	char buf_pid[8];
	char buf_freq[256];

	// to do - need to check new frequncy is valid
	snprintf(buf_pid, sizeof(buf_pid), "%d", getpid());
	snprintf(buf_freq, sizeof(buf_freq), "%d", val * 1000);

	return dbus_cpu_handler(PREDEF_SET_MIN_FREQUENCY, buf_pid, buf_freq);
}

API int deviced_release_cpu_max_frequency()
{
	char buf_pid[8];

	snprintf(buf_pid, sizeof(buf_pid), "%d", getpid());

	return dbus_cpu_handler(PREDEF_RELEASE_MAX_FREQUENCY, buf_pid, "2");
}

API int deviced_release_cpu_min_frequency()
{
	char buf_pid[8];

	snprintf(buf_pid, sizeof(buf_pid), "%d", getpid());

	return dbus_cpu_handler(PREDEF_RELEASE_MIN_FREQUENCY, buf_pid, "2");
}
