/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *	 http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <errno.h>
#include <libgdbus/dbus-system.h>

#include "log.h"
#include "common.h"
#include "dd-usbhost.h"
#include "core/list.h"

#define ARRAY_SIZE(array) (sizeof(array) / sizeof(array[0]))

#define METHOD_OPEN_DEVICE                "OpenDevice"
#define METHOD_REQUEST_STORAGE_INFO_ALL   "StorageInfoAll"
#define METHOD_REQUEST_STORAGE_MOUNT      "StorageMount"
#define METHOD_REQUEST_STORAGE_UNMOUNT    "StorageUnmount"
#define METHOD_REQUEST_STORAGE_FORMAT     "StorageFormat"
#define SIGNAL_NAME_USB_STORAGE_CHANGED   "usb_storage_changed"
#define SIGNAL_NAME_USB_DEVICE_CHANGED    "ChangedDevice"
#define RETRY_MAX 5

union action {
	void (*storage)(char *type, char *path, int mount, void *param);
	void (*device)(struct usbhost_device *device, int state, void *data);
};

struct signal_handler {
	const char *name;
	guint id;
	union action action;
	void *data;
};

static dd_list *handlers;

static int register_signal_handler(const char *path, const char *iface, const char *name,
					GDBusSignalCallback cb, union action action, void *data)
{
	int ret;
	struct signal_handler *handler = NULL;

	handler = (struct signal_handler *)calloc(1, sizeof(struct signal_handler));
	if (!handler) {
		_E("Failed to calloc: No memory");
		ret = -1;
		goto out;
	}
	handler->name = name;
	handler->action = action;
	handler->data = data;
	handler->id = subscribe_dbus_signal(NULL, path, iface, name, cb, NULL, NULL);
	if (!handler->id) {
		_E("Failed to add dbus handler.");
		ret = -1;
		goto out;
	}

	DD_LIST_APPEND(handlers, handler);

	return 0;

out:
	free(handler);
	return ret;
}

static void storage_signal_handler(GDBusConnection  *conn,
					const gchar *sender,
					const gchar *path,
					const gchar *iface,
					const gchar *name,
					GVariant *param,
					gpointer data)
{
	char *type, *syspath;
	int mount;
	dd_list *elem;
	struct signal_handler *handler;

	g_variant_get(param, "(ssi)", &type, &syspath, &mount);

	DD_LIST_FOREACH(handlers, elem, handler) {
		if (strcmp(handler->name, SIGNAL_NAME_USB_STORAGE_CHANGED))
			continue;

		if (handler->action.storage)
			handler->action.storage(type, syspath, mount, handler->data);
	}
	g_free(type);
	g_free(syspath);
}

static void device_signal_handler(GDBusConnection  *conn,
					const gchar *sender,
					const gchar *path,
					const gchar *iface,
					const gchar *name,
					GVariant *param,
					gpointer data)
{
	char *syspath;
	struct usbhost_device device;
	int state;
	struct signal_handler *handler;
	dd_list *element;

	g_variant_get(param, "(isiiiiisss)", &state,
						&syspath,
						&device.baseclass,
						&device.subclass,
						&device.protocol,
						&device.vendorid,
						&device.productid,
						&device.manufacturer,
						&device.product,
						&device.serial);

	strncpy(device.devpath, syspath, PATH_MAX - 1);

	DD_LIST_FOREACH(handlers, element, handler) {
		if (strcmp(handler->name, SIGNAL_NAME_USB_DEVICE_CHANGED))
			continue;

		if (handler->action.device)
			handler->action.device(&device, state, handler->data);
	}
	g_free(syspath);
	g_free(device.manufacturer);
	g_free(device.product);
	g_free(device.serial);
}

API int init_usbhost_signal(void)
{
	_D("Init.");

	return 0;
}

API void deinit_usbhost_signal(void)
{
	dd_list *n, *next;
	struct signal_handler *handler;

	DD_LIST_FOREACH_SAFE(handlers, n, next, handler) {
		DD_LIST_REMOVE(handlers, handler);
		if (handler->id) {
			unsubscribe_dbus_signal(NULL, handler->id);
			handler->id = 0;
		}

		free(handler);
	}

	/*e_dbus_connection_close(conn);
	conn = NULL;

	e_dbus_shutdown();*/
}

API int request_usb_storage_info(void)
{
	return dbus_handle_method_sync(DEVICED_BUS_NAME,
				DEVICED_PATH_USBHOST,
				DEVICED_INTERFACE_USBHOST,
				METHOD_REQUEST_STORAGE_INFO_ALL,
				NULL, NULL);
}

API int register_usb_storage_change_handler(
		void (*storage_changed)(char *type, char *path, int mount, void *param),
		void *data)
{
	union action action;
	int ret;

	if (!storage_changed)
		return -EINVAL;

	action.storage = storage_changed;

	ret = register_signal_handler(DEVICED_PATH_USBHOST,
					DEVICED_INTERFACE_USBHOST,
					SIGNAL_NAME_USB_STORAGE_CHANGED,
					storage_signal_handler,
					action,
					data);
	if (ret < 0)
		return -1;

	return 0;
}

API int register_usb_device_change_handler(
		void (*device_changed)(struct usbhost_device *device, int state, void *data),
		void *data)
{
	union action action;
	int ret;

	if (!device_changed)
		return -EINVAL;

	action.device = device_changed;
	ret = register_signal_handler(DEVICED_PATH_USBHOST,
					DEVICED_INTERFACE_USBHOST,
					SIGNAL_NAME_USB_DEVICE_CHANGED,
					device_signal_handler,
					action,
					data);
	if (ret < 0)
		return -1;

	return 0;
}

API int unregister_usb_storage_change_handler(void)
{
	dd_list *n, *next;
	struct signal_handler *handler;

	DD_LIST_FOREACH_SAFE(handlers, n, next, handler) {
		if (strcmp(handler->name, SIGNAL_NAME_USB_STORAGE_CHANGED))
			continue;
		if (handler->id == 0)
			continue;

		unsubscribe_dbus_signal(NULL, handler->id);
		DD_LIST_REMOVE(handlers, handler);
		free(handler);
	}

	return 0;
}

API int unregister_usb_device_change_handler(
		void (*device_changed)(struct usbhost_device *device, int state, void *data))
{
	dd_list *n, *next;
	struct signal_handler *handler;

	DD_LIST_FOREACH_SAFE(handlers, n, next, handler) {
		if (strcmp(handler->name, SIGNAL_NAME_USB_DEVICE_CHANGED))
			continue;
		if (handler->id == 0)
			continue;
		if (handler->action.device != device_changed)
			continue;

		unsubscribe_dbus_signal(NULL, handler->id);
		DD_LIST_REMOVE(handlers, handler);
		free(handler);
	}

	return -1;
}

API int mount_usb_storage(char *path)
{
	if (!path)
		return -1;

	return dbus_handle_method_sync_var(DEVICED_BUS_NAME,
					DEVICED_PATH_USBHOST,
					DEVICED_INTERFACE_USBHOST,
					METHOD_REQUEST_STORAGE_MOUNT,
					g_variant_new("(s)", path));
}

API int unmount_usb_storage(char *path)
{
	if (!path)
		return -1;

	return dbus_handle_method_sync_var(DEVICED_BUS_NAME,
					DEVICED_PATH_USBHOST,
					DEVICED_INTERFACE_USBHOST,
					METHOD_REQUEST_STORAGE_UNMOUNT,
					g_variant_new("(s)", path));
}

API int format_usb_storage(char *path)
{
	if (!path)
		return -1;

	return dbus_handle_method_sync_var(DEVICED_BUS_NAME,
					DEVICED_PATH_USBHOST,
					DEVICED_INTERFACE_USBHOST,
					METHOD_REQUEST_STORAGE_FORMAT,
					g_variant_new("(s)", path));
}

API int open_usb_device(char *path, int *out_fd)
{
	GVariant *reply = NULL;
	int *fdlist = NULL;
	int ret = 0;
	int size = 0;

	if (!out_fd || !path)
		return -EINVAL;

	reply = dbus_handle_method_with_unix_fd_list_sync_with_reply_var(DEVICED_BUS_NAME,
								DEVICED_PATH_USBHOST,
								DEVICED_INTERFACE_USBHOST,
								METHOD_OPEN_DEVICE,
								g_variant_new("(s)", path),
								NULL,
								0,
								&fdlist,
								&size);
	if (!reply) {
		_E("Unable to open USB device.");
		return -1;
	}
	if (size == 0) {
		_E("Failed to get fd list.");
		return -1;
	}
	if (!dh_get_param_from_var(reply, "(i)", &ret)) {
		_E("Failed to get(%s): no message.", g_variant_get_type_string(reply));
		ret = -1;
		goto out;
	}

	if (ret >= 0)
		*out_fd = fdlist[0];
out:
	if (fdlist)
		free(fdlist);
	if (reply)
		g_variant_unref(reply);

	return ret;
}
