/*
 * deviced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdbool.h>

#include <hw/thermal.h>
#include <libgdbus/dbus-system.h>

#include "apps/apps.h"
#include "core/devices.h"
#include "core/log.h"
#include "core/device-notifier.h"
#include "core/devices.h"
#include "thermal.h"
#include "core/config-parser.h"

static struct thermal_device *thermal_dev;
static int noti;

#define THERMAL_CONF_FILE	"/etc/deviced/temperature.conf"

static struct thermal_config_info thermal_table = {
	.normal   = TEMPERATURE_NORMAL,
	.warning  = TEMPERATURE_WARNING,
	.limit    = TEMPERATURE_LIMIT,
	.shutdown = TEMPERATURE_SHUTDOWN
};

static int power_off(void)
{
	const struct device_ops *power;

	power = find_device("power");
	if (check_default(power))
		return -ENODEV;

	return power->execute("poweroff");
}

static void thermal_remove_noti(void)
{
	int ret = 0;

	if (noti) {
		ret = remove_notification("TempCooldownNotiOff", noti);
		if (ret < 0)
			_E("Failed to remove cooldown noti(%d).", noti);
		else
			noti = 0;
	}
}

static void thermal_remove_popup(void)
{
	int temp;

	temp = launch_system_app(APP_DEFAULT, 2,
		APP_KEY_TYPE, "remove_cooldown_popups");
	if (temp < 0)
		_E("Failed to launch remove_cooldown_popups.");
}

static void thermal_noti_cb(GVariant *var, void *user_data, GError *err)
{
	int id = 0;

	if (!var) {
		if (err)
			_E("%s", err->message);
		return;
	}

	if (!dh_get_param_from_var(var, "(i)", &id)) {
		_E("Failed to get var(%s): no message", g_variant_get_type_string(var));
		goto out;
	}

	noti = id;
	_D("Thermal noti(%d).", noti);

out:
	g_variant_unref(var);
}

static void thermal_add_noti(void)
{
	int ret = 0;

	thermal_remove_noti();
	if (!noti) {
		ret = add_async_notification("TempCooldownNotiOn", thermal_noti_cb, NULL, NULL);
		if (ret < 0)
			_E("Failed to show notification for temperature.");
	}
}

static void broadcast_thermal_state_legacy(const char *action_type)
{
	if (!action_type) {
		_E("Invalid action type.");
		return;
	}

	_D("action_type = %s", action_type);
	dbus_handle_broadcast_dbus_signal_var(DEVICED_PATH_SYSNOTI,
			DEVICED_INTERFACE_SYSNOTI,
			"CoolDownChanged",
			g_variant_new("(s)", action_type));
}

static gboolean thermal_overheat_time_broadcast(void *data)
{
	int ret;
	static int time = OVERHEAT_POWEROFF_TIME;

	/* Transfer Time by Signal */
	ret = dbus_handle_broadcast_dbus_signal_var(POPUP_OVERHEAT_PATH, POPUP_OVERHEAT_INTERFACE,
			SIGNAL_OVERHEAT_TIME, g_variant_new("(i)", time));
	if (ret < 0)
		_E("Failed in updating overheat time.");

	time -= 1;
	if (time < 0) {
		time = OVERHEAT_POWEROFF_TIME;
		ret = launch_system_app(APP_OVERHEAT, 2,
				APP_KEY_TYPE, "remove_overheat_popups");
		if (ret < 0)
			_E("Failed to remove overheat popups.");

		ret = power_off();
		if (ret < 0)
			_E("Failed to power off.");
		return G_SOURCE_REMOVE;
	} else
		return G_SOURCE_CONTINUE;
}

static void thermal_overheat_popup(void)
{
	int ret;
	guint timer;

	ret = launch_system_app(APP_OVERHEAT, 2,
			APP_KEY_TYPE, "overheat");

	if (ret < 0)
		_E("Failed to launch overheat popup.");
	else {
		timer = g_timeout_add_seconds(OVERHEAT_CALLBACK_TIME, thermal_overheat_time_broadcast, NULL);
		if (!timer)
			_E("Failed to set over temp timer.");
	}
}

static void thermal_add_recovery_popup(void)
{
	int temp;

	temp = launch_system_app(APP_DEFAULT, 2,
		APP_KEY_TYPE, "cooled_down");
	if (temp < 0)
		_E("Failed to launch recovery popup.");
}

static void thermal_action(int state)
{
	char *action = NULL;

	switch (state) {
	case NORMAL:
		thermal_remove_noti();
		thermal_add_recovery_popup();
		action = strdup(RELEASE_ACTION);
		break;

	case WARNING:
		_I("State is WARNING.");
		action = strdup(WARNING_ACTION);
		break;

	case LIMIT:
		thermal_remove_popup();
		thermal_add_noti();
		action = strdup(LIMIT_ACTION);
		break;

	case SHUTDOWN:
		thermal_remove_popup();
		thermal_remove_noti();
		thermal_overheat_popup();
		action = strdup(SHUTDOWN_ACTION);
		break;
	}

	broadcast_thermal_state_legacy(action);
	free(action);
}

/* Action Table
--------------------------------------------------
Current		Previous	Action
--------------------------------------------------
NORMAL		NORMAL		X
		WARNING		X
		LIMIT		ReleaseAction
		SHUTDOWN	X
--------------------------------------------------
WARNING		NORMAL		print log
		WARNING		X
		LIMIT		X
		SHUTDOWN	X
--------------------------------------------------
LIMIT		NORMAL		LimitAction
		WARNING		LimitAction
		LIMIT		X
		SHUTDOWN	X
--------------------------------------
SHUTDOWN	NORMAL		ShutdownAction
		WARNING		ShutdownAction
		LIMIT		ShutdownAction
		SHUTDOWN	X
--------------------------------------------------
*/
static bool is_action_needed(int old_state, int new_state)
{
	bool ret = false;

	if (new_state == NORMAL) {
		if (old_state == WARNING || old_state == LIMIT)
			ret = true;
	} else if (new_state == WARNING) {
		if (old_state == NORMAL)
			ret = true;
	} else if (new_state == LIMIT) {
		if (old_state == NORMAL || old_state == WARNING)
			ret = true;
	} else if (new_state == SHUTDOWN) {
		if (old_state != SHUTDOWN)
			ret = true;
	}

	return ret;
}

static int thermal_get_temp_enum(int temp)
{
	int ret = -1;

	if (temp <= thermal_table.normal)
		ret = NORMAL;
	else if (temp >= thermal_table.shutdown)
		ret = SHUTDOWN;
	else if (temp >= thermal_table.limit)
		ret = LIMIT;
	else if (temp >= thermal_table.warning)
		ret = WARNING;

	return ret;
}

static int thermal_get_average(int temp)
{
	static int buf[SAMPLE_CNT] = {25, 25, 25, 25};
	static int index = 0;
	int i, avg_temp = 0;

	buf[index++] = temp;
	index %= SAMPLE_CNT;

	for (i = 0; i < SAMPLE_CNT; i++)
		avg_temp += buf[i];

	avg_temp /= SAMPLE_CNT;
	return avg_temp;
}

static void thermal_handler(struct thermal_info *info, void *data)
{
	static int old_state = NORMAL;
	int avg_temp, new_state;
	bool action_needed = false;

	avg_temp = thermal_get_average(info->temp);
	new_state = thermal_get_temp_enum(avg_temp);
	if (new_state == -1) {
		_I("Keep current state: %d", old_state);
		return;
	}

	action_needed = is_action_needed(old_state, new_state);
	_I("Action=%d old_state=%d new_state=%d", action_needed, old_state, new_state);
	if (action_needed)
		thermal_action(new_state);

	old_state = new_state;
}

static int thermal_service_start(void)
{
	int ret;

	if (!thermal_dev) {
		_E("Thermal service is not supported.");
		return -ENOTSUP;
	}

	ret = thermal_dev->register_changed_event(thermal_handler, NULL);
	if (ret < 0)
		_E("Failed to register thermal event: %d", ret);

	return ret;
}

static int thermal_service_stop(void)
{
	int ret;

	if (!thermal_dev) {
		_E("Thermal service is not supported.");
		return -ENOTSUP;
	}

	ret = thermal_dev->unregister_changed_event(thermal_handler);
	if (ret < 0)
		_E("Failed to unregister thermal event: %d", ret);

	return ret;
}

static GVariant *dbus_get_temperature(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	struct thermal_info info;
	int index, ret = 0;

	g_variant_get(param, "(i)", &index);

	if (!thermal_dev || !thermal_dev->get_info) {
		ret = -ENODEV;
		goto exit;
	}

	ret = thermal_dev->get_info((device_thermal_e)index, &info);
	if (!ret)
		ret = info.temp;

exit:
	return g_variant_new("(i)", ret);
}

static const dbus_method_s dbus_methods[] = {
	{ "GetTemperature", "i", "i", dbus_get_temperature},
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_TEMPERATURE,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static int booting_done(void *data)
{
	static int done;

	if (data == NULL)
		return done;
	done = *(int *)data;
	if (done == 0)
		return done;

	_I("Booting done.");
	thermal_service_start();

	return done;
}

static int load_config(struct parse_result *result, void *user_data)
{
	struct thermal_config_info *info = user_data;
	char *name;
	char *value;

	_D("Load config. section=%s name=%s value=%s", result->section, result->name, result->value);

	if (!info)
		return -EINVAL;

	if (!MATCH(result->section, "THERMAL"))
		return -EINVAL;

	name = result->name;
	value = result->value;
	if (MATCH(name, "Normal"))
		info->normal = atoi(value);
	else if (MATCH(name, "Warning"))
		info->warning = atoi(value);
	else if (MATCH(name, "Limit"))
		info->limit = atoi(value);
	else if (MATCH(name, "Shutdown"))
		info->shutdown = atoi(value);

	return 0;
}

static void thermal_init(void *data)
{
	int ret;

	/* load thermal configuration file */
	ret = config_parse(THERMAL_CONF_FILE, load_config, &thermal_table);
	if (ret < 0)
		_E("Failed to load power off config: %d", ret);

	_I("Temperature conf normal=%d warning=%d limit=%d shutdown=%d", thermal_table.normal, thermal_table.warning,
		thermal_table.limit, thermal_table.shutdown);

	ret = register_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
	if (ret < 0)
		_E("Failed to register booting done notifier.");

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_TEMPERATURE, &dbus_interface);
	if (ret < 0)
		_E("Failed to init dbus method: %d", ret);
}

static void thermal_exit(void *data)
{
	int ret;

	ret = unregister_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
	if (ret < 0)
		_E("Failed to unregister booting done notifier.");

	thermal_service_stop();
}

static int thermal_probe(void *data)
{
	struct hw_info *info;
	int ret;

	if (thermal_dev)
		return 0;

	ret = hw_get_info(THERMAL_HARDWARE_DEVICE_ID,
			(const struct hw_info **)&info);
	if (ret < 0) {
		_E("There is no HAL for Thermal.");
		return ret;
	}

	if (!info->open) {
		_E("Failed to open thermal device: open(NULL)");
		return -ENODEV;
	}

	ret = info->open(info, NULL, (struct hw_common**)&thermal_dev);
	if (ret < 0) {
		_E("Failed to get thermal device structure: %d", ret);
		return ret;
	}

	if (!thermal_dev ||
		!thermal_dev->get_info ||
		!thermal_dev->register_changed_event ||
		!thermal_dev->unregister_changed_event) {
		_E("Thermal HAL does not support event handler.");
		return -ENODEV;
	}

	_I("Thermal device structure load success.");
	return 0;
}

static const struct device_ops thermal_ops = {
	.name     = "thermal",
	.probe    = thermal_probe,
	.init     = thermal_init,
	.exit     = thermal_exit,
};

DEVICE_OPS_REGISTER(&thermal_ops)
