/*
 * deviced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdbool.h>
#include <device-node.h>
#include <vconf.h>
#include <hw/touchsensitivity.h>
#include <libgdbus/dbus-system.h>

#include "core/devices.h"
#include "core/common.h"
#include "core/log.h"
#include "core/device-notifier.h"


#ifndef VCONFKEY_SETAPPL_TOUCH_SENSITIVITY
#define VCONFKEY_SETAPPL_TOUCH_SENSITIVITY "db/setting/enhance_touch_sensitivity"
#endif

#define DISABLE_SENSITIVITY_SIGNAL "DisableSensitivity"

#define TOUCH_SENSITIVITY_DISABLE_INTERVAL 1

#ifndef TOUCHSENSITIVITY_GLOVE_MODE_OFF
#define TOUCHSENSITIVITY_GLOVE_MODE_OFF 0
#endif
#ifndef TOUCHSENSITIVITY_GLOVE_MODE_ON
#define TOUCHSENSITIVITY_GLOVE_MODE_ON  1
#endif

static guint disable_timer;

static struct touchsensitivity_device *touchsensitivity_dev;

static int check_dev(void)
{
	if (!touchsensitivity_dev ||
	    !touchsensitivity_dev->glove_mode_set_state ||
	    !touchsensitivity_dev->glove_mode_get_state)
		return -ENODEV;
	return 0;
}

static gboolean disable_sensitivity(void *data)
{
	int ret;
	int status;

	if (check_dev() < 0) {
		_E("Touch sensitivity HAL does not exist.");
		return G_SOURCE_REMOVE;
	}

	if (vconf_get_bool(VCONFKEY_SETAPPL_TOUCH_SENSITIVITY, &status) != 0 ||
	    status == TOUCHSENSITIVITY_GLOVE_MODE_ON)
		return G_SOURCE_REMOVE;

	ret = touchsensitivity_dev->glove_mode_set_state(TOUCHSENSITIVITY_GLOVE_MODE_OFF);
	_I("Change auto touch sensitivity disable: %s", ret ? "fail" : "set");
	return G_SOURCE_REMOVE;
}

static void disable_touch_sensitivity_handler(GDBusConnection  *conn,
	const gchar      *sender,
	const gchar      *path,
	const gchar      *iface,
	const gchar      *name,
	GVariant         *param,
	gpointer          data)
{
	disable_timer = g_timeout_add_seconds(TOUCH_SENSITIVITY_DISABLE_INTERVAL, disable_sensitivity, NULL);
	if (disable_timer == 0)
		_E("Failed to add battery init timer during booting.");
}

static GVariant *get_touch_sensitivity(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int val = -1, ret;

	if (check_dev() < 0) {
		_E("Touch sensitivity HAL does not exist.");
		goto out;
	}
	ret = touchsensitivity_dev->glove_mode_get_state(&val);
	_I("Touch sensitivity(%d): %s", val, ret ? "fail" : "get");
out:
	return g_variant_new("(i)", val);
}

static GVariant *set_touch_sensitivity(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int val = -1, ret = -ESRCH;

	g_variant_get(param, "(i)", &val);

	if (disable_timer) {
		g_source_remove(disable_timer);
		disable_timer = 0;
	}
	if (check_dev() < 0) {
		_E("Touch sensitivity HAL does not exist.");
		ret = -EINVAL;
		goto error;
	}
	ret = touchsensitivity_dev->glove_mode_set_state(val);

error:
	_I("Touch sensitivity(%d): %s", val, ret ? "fail" : "set");

	return g_variant_new("(i)", ret);
}

static const dbus_method_s dbus_methods[] = {
	{ "GetSensitivity", NULL,  "i", get_touch_sensitivity },
	{ "SetSensitivity",  "i",  "i", set_touch_sensitivity },
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_TOUCH,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static int booting_done(void *data)
{
	static int done;
	int ret;
	int status = TOUCHSENSITIVITY_GLOVE_MODE_OFF;

	if (data == NULL)
		goto out;

	done = *(int *)data;

	if (vconf_get_bool(VCONFKEY_SETAPPL_TOUCH_SENSITIVITY, &status) != 0 ||
	    status == TOUCHSENSITIVITY_GLOVE_MODE_OFF)
		goto out;

	if (check_dev() < 0) {
		_E("Touch sensitivity HAL does not exist.");
		goto out;
	}
	ret = touchsensitivity_dev->glove_mode_set_state(TOUCHSENSITIVITY_GLOVE_MODE_ON);
	_I("Change auto touch sensitivity enable: %s", ret ? "fail" : "set");
	unregister_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
out:
	return done;
}

static void sensitivity_init(void *data)
{
	int ret;
	struct hw_info *info;

	if (touchsensitivity_dev)
		return;

	ret = hw_get_info(TOUCHSENSITIVITY_HARDWARE_DEVICE_ID,
			(const struct hw_info **)&info);
	if (ret < 0) {
		_E("Failed to load touch sensitivity shared library: %d", ret);
		return;
	}

	if (!info->open) {
		_E("Failed to open touch sensitivity device: open(NULL)");
		return;
	}

	ret = info->open(info, NULL, (struct hw_common **)&touchsensitivity_dev);
	if (ret < 0) {
		_E("Failed to get touch sensitivity device structure: %d", ret);
		return;
	}

	_I("Touch sensitivity device structure load success.");

	/* init dbus interface */
	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_TOUCH, &dbus_interface);
	if (ret < 0)
		_E("Failed to init dbus method: %d", ret);

	subscribe_dbus_signal(NULL,
		DEVICED_PATH_TOUCH,
		DEVICED_INTERFACE_TOUCH,
		DISABLE_SENSITIVITY_SIGNAL,
		disable_touch_sensitivity_handler,
		NULL, NULL);

	register_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
}

static const struct device_ops sensitivity_device_ops = {
	.name     = "sensitivity",
	.init     = sensitivity_init,
};

DEVICE_OPS_REGISTER(&sensitivity_device_ops)

